from datetime import datetime

from pydantic import BaseModel


class SiteGroupResponse(BaseModel):
    """Response model for SiteGroup - includes basic summary information"""

    site_group_id: int
    name: str
    created_time: datetime
    changed_time: datetime
    total_sites: int  # How many sites are linked to this group


class SiteGroupPageResponse(BaseModel):
    """Represents a paginated response of SiteGroup"""

    total_count: int  # The total number of groups (independent of this page of results)
    limit: int  # The maximum number of groups that could've been returned (the limit set by the query)
    start: int  # The number of groups that have been skipped as part of this query (the start set by the query)
    groups: list[SiteGroupResponse]  # The group models in this page



from datetime import datetime
from typing import Generic, Optional, TypeVar

from pydantic import BaseModel

from envoy_schema.admin.schema.doe import DynamicOperatingEnvelopeResponse
from envoy_schema.admin.schema.pricing import TariffGeneratedRateResponse
from envoy_schema.admin.schema.site import SiteResponse


class ArchiveBase(BaseModel):
    """The common properties for all archive responses"""

    archive_id: int  # Unique PK for identifying an individual archive record
    archive_time: datetime  # When was this archive record created
    deleted_time: Optional[datetime]  # When was this archive record marked as deleted (if None - not deleted)


class ArchiveDynamicOperatingEnvelopeResponse(ArchiveBase, DynamicOperatingEnvelopeResponse):
    """Archived version of DynamicOperatingEnvelopeResponse"""

    pass


class ArchiveTariffGeneratedRateResponse(ArchiveBase, TariffGeneratedRateResponse):
    """Archived version of TariffGeneratedRateResponse"""

    pass


class ArchiveSiteResponse(ArchiveBase, SiteResponse):
    """Archived version of SiteResponse"""

    pass


ArchiveType = TypeVar(
    "ArchiveType", ArchiveDynamicOperatingEnvelopeResponse, ArchiveTariffGeneratedRateResponse, ArchiveSiteResponse
)


class ArchivePageResponse(BaseModel, Generic[ArchiveType]):
    """Represents a paginated response of archive entities"""

    total_count: int  # The total number of entities (independent of this page of results)
    limit: int  # The maximum number of entities that could've been returned (the limit set by the query)
    start: int  # The number of entities that have been skipped as part of this query (the start set by the query)
    period_start: datetime  # The period_start parameter from the original request
    period_end: datetime  # The period_end parameter from the original request
    entities: list[ArchiveType]  # The entity models in this paged response



from datetime import datetime
from decimal import Decimal
from typing import Optional

from pydantic import BaseModel

from envoy_schema.server.schema.sep2.der import (
    AbnormalCategoryType,
    AlarmStatusType,
    ConnectStatusType,
    DERControlType,
    DERType,
    DOESupportedMode,
    InverterStatusType,
    LocalControlModeStatusType,
    NormalCategoryType,
    OperationalModeStatusType,
)
from envoy_schema.server.schema.sep2.types import DeviceCategory


class DERConfiguration(BaseModel):
    """Represents a combination of the sep2 DERSettings + DERCapability. The values here will use the current setting
    value (if specified) otherwise they will fallback to the nameplate rating (if available)"""

    # Mandatory values
    created_time: datetime
    changed_time: datetime
    modes_supported: DERControlType
    type: DERType
    doe_modes_supported: DOESupportedMode
    max_w: Decimal  # Max continuous active power in watts

    # Optional values
    abnormal_category: Optional[AbnormalCategoryType]
    normal_category: Optional[NormalCategoryType]
    max_a: Optional[Decimal]  # Max continuous AC current capability in Amperes
    max_ah: Optional[Decimal]  # Usable energy storage in AmpHours
    max_charge_rate_va: Optional[Decimal]
    max_charge_rate_w: Optional[Decimal]
    max_discharge_rate_va: Optional[Decimal]
    max_discharge_rate_w: Optional[Decimal]
    max_v: Optional[Decimal]
    min_v: Optional[Decimal]
    max_va: Optional[Decimal]
    max_var: Optional[Decimal]  # Max reactive power delivered by the DER in VAR
    max_var_neg: Optional[Decimal]  # Max reactive power receivable by the DER in VAR. Defaults to -'ve max_var
    max_wh: Optional[Decimal]
    v_nom: Optional[Decimal]  # Nominal AC voltage


class DERAvailability(BaseModel):
    """ "Represents the current availability values associated with a Site's DER. Typically used for communicating
    the current snapshot of DER energy held in reserve"""

    # Mandatory values
    created_time: datetime
    changed_time: datetime

    # Optional values
    availability_duration_sec: Optional[int]
    max_charge_duration_sec: Optional[int]
    reserved_charge_percent: Optional[Decimal]
    reserved_deliver_percent: Optional[Decimal]
    estimated_var_avail: Optional[Decimal]
    estimated_w_avail: Optional[Decimal]


class DERStatus(BaseModel):
    """Represents the current status values associated with a Site's DER. Typically used for communicating
    the current snapshot of DER status"""

    # Mandatory values
    created_time: datetime
    changed_time: datetime

    # Optional values
    alarm_status: Optional[AlarmStatusType]
    generator_connect_status: Optional[ConnectStatusType]
    generator_connect_status_time: Optional[datetime]
    inverter_status: Optional[InverterStatusType]
    inverter_status_time: Optional[datetime]
    local_control_mode_status: Optional[LocalControlModeStatusType]
    local_control_mode_status_time: Optional[datetime]
    manufacturer_status: Optional[str]
    manufacturer_status_time: Optional[datetime]
    operational_mode_status: Optional[OperationalModeStatusType]
    operational_mode_status_time: Optional[datetime]


class SiteGroup(BaseModel):
    """Represents a named group that a site might belong to"""

    site_group_id: int
    name: str
    created_time: datetime
    changed_time: datetime


class SiteResponse(BaseModel):
    """Response model for Site - includes the common details"""

    aggregator_id: int
    site_id: int
    nmi: Optional[str]
    timezone_id: str
    created_time: datetime
    changed_time: datetime
    lfdi: str
    sfdi: int
    device_category: DeviceCategory

    groups: list[SiteGroup]

    der_config: Optional[DERConfiguration]  # Metadata about site DER - no guarantee on availability
    der_availability: Optional[DERAvailability]  # Metadata about site DER - no guarantee on availability
    der_status: Optional[DERStatus]  # Metadata about site DER - no guarantee on availability


class SitePageResponse(BaseModel):
    """Represents a paginated response of Site"""

    total_count: int  # The total number of sites (independent of this page of results)
    limit: int  # The maximum number of sites that could've been returned (the limit set by the query)
    start: int  # The number of sites that have been skipped as part of this query (the start set by the query)
    group: Optional[str]  # The "group" filter set by the query (if any)
    after: Optional[datetime]  # The "after" filter set by the query (if any)
    sites: list[SiteResponse]  # The site models in this page



from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema import uri
from envoy_schema.server.schema.sep2.function_set_assignments import FunctionSetAssignmentsBase
from envoy_schema.server.schema.sep2.identification import ListLink
from envoy_schema.server.schema.sep2.types import DEFAULT_POLLRATE_SECONDS


class DeviceCapabilityResponse(FunctionSetAssignmentsBase, tag="DeviceCapability"):
    href: str = attr(default=uri.DeviceCapabilityUri)
    pollRate: Optional[int] = attr(default=DEFAULT_POLLRATE_SECONDS)  # recommended client pollrate in seconds

    # (0..1) Link
    # Not supported at this time
    # SelfDeviceLink: Optional[Link] = element(default=None)

    # (0..1) ListLink
    EndDeviceListLink: Optional[ListLink] = element(default=None)
    MirrorUsagePointListLink: Optional[ListLink] = element(default=None)



from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2.identification import (
    IdentifiedObject,
    Link,
    ListLink,
    Resource,
    SubscribableList,
    SubscribableResource,
)
from envoy_schema.server.schema.sep2.types import DEFAULT_POLLRATE_SECONDS


class FunctionSetAssignmentsBase(Resource):
    # Optional (0..1) Links and ListLinks
    CustomerAccountListLink: Optional[ListLink] = element(default=None)
    DemandResponseProgramListLink: Optional[ListLink] = element(default=None)
    DERProgramListLink: Optional[ListLink] = element(default=None)
    FileListLink: Optional[ListLink] = element(default=None)
    MessagingProgramListLink: Optional[ListLink] = element(default=None)
    PrepaymentListLink: Optional[ListLink] = element(default=None)
    ResponseSetListLink: Optional[ListLink] = element(default=None)
    TariffProfileListLink: Optional[ListLink] = element(default=None)
    TimeLink: Optional[Link] = element(default=None)
    UsagePointListLink: Optional[ListLink] = element(default=None)


# The SEP2 standard doesn't explicitly state that FunctionSetAssignments derives from
# IdentifiedObject nor SubscribableResource. However the fields present on FunctionSetAssignments
# matches those present in IdentifiedObject and SubscribableResource so we have decided to inherit from these
# in addition to explicitly stated parent class, namely, FunctionSetAssignmentsBase
class FunctionSetAssignmentsResponse(
    IdentifiedObject, FunctionSetAssignmentsBase, SubscribableResource, tag="FunctionSetAssignments"
):
    pass


class FunctionSetAssignmentsListResponse(SubscribableList, tag="FunctionSetAssignmentsList"):

    FunctionSetAssignments: list[FunctionSetAssignmentsResponse] = element(default_factory=list)
    pollRate: Optional[int] = attr(default=DEFAULT_POLLRATE_SECONDS)  # recommended client pollrate in seconds



from urllib.parse import urlparse

from pydantic import AfterValidator
from typing_extensions import Annotated


def validate_String6(v: str):
    if len(v) > 6:
        raise ValueError("String6 max length of 6.")
    return v


def validate_String32(v: str):
    if len(v) > 32:
        raise ValueError("String32 max length of 32.")
    return v


def validate_String192(v: str):
    if len(v) > 192:
        raise ValueError("String192 max length of 192.")
    return v


def validate_HexBinary8(v: str):
    if len(v) > 2:
        raise ValueError("HexBinary8 max length of 2.")
    return v


def validate_HexBinary16(v: str):
    if len(v) > 4:
        raise ValueError("HexBinary16 max length of 4.")
    return v


def validate_HexBinary32(v: str):
    if len(v) > 8:
        raise ValueError("HexBinary32 max length of 8.")
    return v


def validate_HexBinary48(v: str):
    if len(v) > 12:
        raise ValueError("HexBinary48 max length of 12.")
    return v


def validate_HexBinary64(v: str):
    if len(v) > 16:
        raise ValueError("HexBinary64 max length of 16.")
    return v


def validate_HexBinary128(v: str):
    if len(v) > 32:
        raise ValueError("HexBinary128 max length of 32.")
    return v


def validate_HexBinary160(v: str):
    if len(v) > 40:
        raise ValueError("HexBinary160 max length of 40.")
    return v


def validate_LocalAbsoluteUri(v: str):
    """Only does a cursory check that a URI looks like a local absolute URI eg: /edev/123/cp"""
    v = v.strip()
    if len(v) > 4096:
        raise ValueError("LocalUri length has a max of 4096")

    parsed = urlparse(v)
    if parsed.scheme or parsed.netloc:
        raise ValueError("LocalUri should not include a scheme or host")

    if not v.startswith("/"):
        raise ValueError("LocalUri should be an absolute URI")

    return v


def validate_HttpUri(v: str):
    """Only does a cursory check that a URI looks like a remote server HTTP(S) query eg: https://example.com:123/hook"""

    v = v.strip()
    if len(v) > 4096:
        raise ValueError("HttpUri length has a max of 4096")

    parsed = urlparse(v)
    if parsed.scheme != "https" and parsed.scheme != "http":
        raise ValueError("HttpUri should have a http or https scheme")

    if len(parsed.netloc) < 3:
        raise ValueError("HttpUri requires a remote host")

    if not parsed.path.startswith("/"):
        raise ValueError("HttpUri should be an absolute path")

    return v


String6 = Annotated[str, AfterValidator(validate_String6)]
String32 = Annotated[str, AfterValidator(validate_String32)]
String192 = Annotated[str, AfterValidator(validate_String192)]

HexBinary8 = Annotated[str, AfterValidator(validate_HexBinary8)]
HexBinary16 = Annotated[str, AfterValidator(validate_HexBinary16)]
HexBinary32 = Annotated[str, AfterValidator(validate_HexBinary32)]
HexBinary48 = Annotated[str, AfterValidator(validate_HexBinary48)]
HexBinary64 = Annotated[str, AfterValidator(validate_HexBinary64)]
HexBinary128 = Annotated[str, AfterValidator(validate_HexBinary128)]
HexBinary160 = Annotated[str, AfterValidator(validate_HexBinary160)]


LocalAbsoluteUri = Annotated[str, AfterValidator(validate_LocalAbsoluteUri)]
HttpUri = Annotated[str, AfterValidator(validate_HttpUri)]



from enum import IntEnum, IntFlag, auto
from functools import reduce

from pydantic_xml import element

from envoy_schema.server.schema.sep2 import base

# Version SHALL indicate a distinct identifier for each revision of an IdentifiedObject. If not specified, a default
# version of "0" (initial version) SHALL be assumed. Upon modification of any IdentifiedObject, the mRID SHALL remain
# the same, but the version SHALL be incremented. Servers MAY NOT modify objects that they did not create, unless they
# were notified of the change from the entity controlling the object's PEN.
VersionType = int  # Expected to be uint16


class AccumulationBehaviourType(IntEnum):
    """sep2 AccumulationBehaviourType type. All other values are reserved"""

    NOT_APPLICABLE = 0
    CUMULATIVE = 3
    DELTA_DATA = 4
    INDICATING = 6
    SUMMATION = 9
    INSTANTANEOUS = 12


class UomType(IntEnum):
    """Described in sep2 as:

    The following values are recommended values sourced from the unit of measure enumeration in IEC 61968-9 [61968].
    Other values from the unit of measure enumeration in IEC 61968-9 [61968] MAY be used."""

    NOT_APPLICABLE = 0
    CURRENT_AMPERES = 5
    TEMPERATURE_KELVIN = 6
    TEMPERATURE_CELSIUS = 23
    VOLTAGE = 29
    JOULES = 31
    FREQUENCY_HZ = 33
    REAL_POWER_WATT = 38
    VOLUME_CUBIC_METRE = 42
    APPARENT_POWER_VA = 61
    REACTIVE_POWER_VAR = 63
    DISPLACEMENT_POWER_FACTOR_COSTHETA = 65
    VOLTS_SQUARED = 67
    AMPERES_SQUARED = 69
    APPARENT_ENERGY_VAH = 71
    REAL_ENERGY_WATT_HOURS = 72
    REACTIVE_ENERGY_VARH = 73
    AVAILABLE_CHARGE_AMPERE_HOURS = 106
    VOLUME_CUBIC_FEET = 119
    VOLUME_CUBIC_FEET_PER_HOUR = 122
    VOLUME_CUBIC_METRE_PER_HOUR = 125
    VOLUME_US_GALLON = 128
    VOLUME_US_GALLON_PER_HOUR = 129
    VOLUME_IMPERIAL_GALLON = 130
    VOLUME_IMPERIAL_GALLON_PER_HOUR = 131
    BRITISH_THERMAL_UNIT = 132
    BRITISH_THERMAL_UNIT_PER_HOUR = 133
    VOLUME_LITER = 134
    VOLUME_LITER_PER_HOUR = 137


class CommodityType(IntEnum):
    """All other values reserved"""

    NOT_APPLICABLE = 0
    ELECTRICITY_SECONDARY_METERED_VALUE = 1
    ELECTRICITY_PRIMARY_METERED_VALUE = 2
    AIR = 4
    NATURAL_GAS = 7
    PROPANE = 8
    POTABLE_WATER = 9
    STEAM = 10
    WASTE_WATER = 11
    HEATING_FLUID = 12
    COOLING_FLUID = 13


class DataQualifierType(IntEnum):
    """All other values reserved"""

    NOT_APPLICABLE = 0
    AVERAGE = 2
    MAXIMUM = 8
    MINIMUM = 9
    STANDARD = 12
    STD_DEVIATION_OF_POPULATION = 29
    STD_DEVIATION_OF_SAMPLE = 30


class FlowDirectionType(IntEnum):
    """All other values reserved"""

    NOT_APPLICABLE = 0
    FORWARD = 1  # delivered to customer
    REVERSE = 19  # received from customer


class KindType(IntEnum):
    """All other values reserved"""

    NOT_APPLICABLE = 0
    CURRENCY = 3
    DEMAND = 8
    ENERGY = 12
    POWER = 37


class PhaseCode(IntEnum):
    """All other values reserved"""

    NOT_APPLICABLE = 0
    PHASE_C_S2 = 32
    PHASE_CN_S2N = 33
    PHASE_CA = 40
    PHASE_B = 64
    PHASE_BN = 65
    PHASE_BC = 66
    PHASE_A_S1 = 128
    PHASE_AN_S1N = 129
    PHASE_AB = 132
    PHASE_ABC = 224


class TOUType(IntEnum):
    """All other values reserved"""

    NOT_APPLICABLE = 0
    TOU_A = 1
    TOU_B = 2
    TOU_C = 3
    TOU_D = 4
    TOU_E = 5
    TOU_F = 6
    TOU_G = 7
    TOU_H = 8
    TOU_I = 9
    TOU_J = 10
    TOU_K = 11
    TOU_L = 12
    TOU_M = 13
    TOU_N = 14
    TOU_O = 15


class ConsumptionBlockType(IntEnum):
    """All other values reserved"""

    NOT_APPLICABLE = 0
    BLOCK_1 = 1
    BLOCK_2 = 2
    BLOCK_3 = 3
    BLOCK_4 = 4
    BLOCK_5 = 5
    BLOCK_6 = 6
    BLOCK_7 = 7
    BLOCK_8 = 8
    BLOCK_9 = 9
    BLOCK_10 = 10
    BLOCK_11 = 11
    BLOCK_12 = 12
    BLOCK_13 = 13
    BLOCK_14 = 14
    BLOCK_15 = 15
    BLOCK_16 = 16


class SubscribableType(IntEnum):
    resource_does_not_support_subscriptions = 0
    resource_supports_non_conditional_subscriptions = 1
    resource_supports_conditional_subscriptions = 2
    resource_supports_both_conditional_and_non_conditional_subscriptions = 3


class UnitValueType(base.BaseXmlModelWithNS):
    """Type for specification of a specific value, with units and power of ten multiplier."""

    multiplier: int = element()
    unit: UomType = element()
    value: int = element()


DEFAULT_POLLRATE_SECONDS: int = 900  # pollrate default as defined by sep2 - Ends up being 15 minutes everywhere


class DeviceCategory(IntFlag):
    """DeviceCategory is a series of bit flags describing a category of EndDevice. Described in sep2"""

    PROGRAMMABLE_COMMUNICATING_THERMOSTAT = auto()
    STRIP_HEATERS = auto()
    BASEBOARD_HEATERS = auto()
    WATER_HEATER = auto()
    POOL_PUMP = auto()
    SAUNA = auto()
    HOT_TUB = auto()
    SMART_APPLIANCE = auto()
    IRRIGATION_PUMP = auto()
    MANAGED_COMMERCIAL_AND_INDUSTRIAL_LOADS = auto()
    SIMPLE_MISC_LOADS = auto()
    EXTERIOR_LIGHTING = auto()
    INTERIOR_LIGHTING = auto()
    LOAD_CONTROL_SWITCH = auto()
    ENERGY_MANAGEMENT_SYSTEM = auto()
    SMART_ENERGY_MODULE = auto()
    ELECTRIC_VEHICLE = auto()
    ELECTRIC_VEHICLE_SUPPLY_EQUIPMENT = auto()
    VIRTUAL_OR_MIXED_DER = auto()
    RECIPROCATING_ENGINE = auto()
    FUEL_CELL = auto()
    PHOTOVOLTAIC_SYSTEM = auto()
    COMBINED_HEAT_AND_POWER = auto()
    COMBINED_PV_AND_STORAGE = auto()
    OTHER_GENERATION_SYSTEM = auto()
    OTHER_STORAGE_SYSTEM = auto()


# The combination of ALL DeviceCategory bit flags
DEVICE_CATEGORY_ALL_SET: DeviceCategory = reduce(lambda a, b: a | b, DeviceCategory)  # type: ignore # py311 issue

# Time is a signed 64 bit value representing the number of seconds since 0 hours, 0 minutes, 0 seconds, on the 1st of
# January, 1970, in UTC, not counting leap seconds
TimeType = int

# A signed time offset, typically applied to a Time value, expressed in seconds.
TimeOffsetType = int


class TimeQualityType(IntEnum):
    authoritative_source = 3
    level_3_source = 4
    level_4_source = 5
    level_5_source = 6
    intentionally_uncoordinated = 7


class DateTimeIntervalType(base.BaseXmlModelWithNS, tag="DateTimeInterval"):
    duration: int = element()
    start: TimeType = element()


class CurrencyCode(IntEnum):
    """Non exhaustive set of numerical ISO 4217 currency codes. Described in sep2 / ISO 4217"""

    NOT_APPLICABLE = 0
    AUSTRALIAN_DOLLAR = 36
    CANADIAN_DOLLAR = 124
    US_DOLLAR = 840
    EURO = 978


class PrimacyType(IntEnum):
    """Values possible for indication of Primary provider. Described in sep2.

    It's worth noting that Values 3-64 are reserved, values 65-191 are user definable and 192-255 are also reserved

    Lower numbers indicate higher priority"""

    IN_HOME_ENERGY_MANAGEMENT_SYSTEM = 0
    CONTRACTED_PREMISES_SERVICE_PROVIDER = 1


class ServiceKind(IntEnum):
    """sep2 ServiceKind type. All other values are reserved"""

    ELECTRICITY = 0
    GAS = 1
    WATER = 2
    TIME = 3
    PRESSURE = 4
    HEAT = 5
    COOLING = 6


class RoleFlagsType(IntFlag):
    """Specifies the roles that apply to a usage point. Described in sep2. Other bits reserved"""

    NONE = 0
    IS_MIRROR = auto()
    IS_PREMISES_AGGREGATION_POINT = auto()
    IS_PEV = auto()
    IS_DER = auto()
    IS_REVENUE_QUALITY = auto()
    IS_DC = auto()
    IS_SUBMETER = auto()


class ReasonCodeType(IntEnum):
    invalid_request_format = 0
    invalid_request_values = 1
    resource_limit_reached = 2
    conditional_subscription_field_not_supported = 3
    maximum_request_frequency_exceeded = 4

    # Custom values outside sep2 specification
    internal_error = 16384  # Unspecified error due to an issue with some internal logic/system


# A signed time offset, typically applied to a Time value, expressed in seconds, with range -3600 to 3600.
OneHourRangeType = int

# Used for signed percentages, specified in hundredths of a percent, -10000 - 10000. (10000 = 100%)
SignedPerCent = int

# Used for percentages, specified in hundredths of a percent, 0 - 10000. (10000 = 100%)
PerCent = int

# 6 digit unsigned decimal integer 0 -> 999,999 (last digit being a checksum digit)
PINType = int


class DERUnitRefType(IntEnum):
    """Specifies context for interpreting percent values:. All other values are reserved"""

    NOT_APPLICABLE = 0
    PERC_SET_MAX_W = 1
    PERC_SET_MAX_VAR = 2
    PERC_STAT_VAR_AVAIL = 3
    PERC_SET_EFFECTIVE_V = 4
    PERC_SET_MAX_CHARGE_RATE_W = 5
    PERC_SET_MAX_DISCHARGE_RATE_W = 6
    PERC_STAT_W_AVAIL = 7


class QualityFlagsType(IntFlag):
    """There is no corresponding definition of this within sep2 - it's instead outlined on the HexBinary16 definition
    of ReadingBase.qualityFlags"""

    NONE = 0
    VALID = auto()  # data that has gone through all required validation checks is now considered valid / verified
    MANUALLY_EDITED = auto()  # manually edited: Replaced or approved by a human
    ESTIMATED_BY_DAY = auto()  # value was replaced by a machine computed value based on analysis of historical data
    ESTIMATED_BY_LINEAR = auto()  # data value was computed using linear interp based on the readings before / after it
    QUESTIONABLE = auto()  # data that has failed one or more checks
    DERIVED = auto()  # data that has been calculated (using logic or mathematical operations), not measured directly
    FORECAST = auto()  # data that has been calculated as a projection or forecast of future readings



from enum import IntEnum
from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2 import primitive_types, types
from envoy_schema.server.schema.sep2.identification import Resource, SubscribableList


class FunctionSetIdentifier(IntEnum):
    """The various FunctionSetIdentifiers defined in sep2. All other values are reserved."""

    GENERAL = 0
    PUBLISH_AND_SUBSCRIBE = 1
    END_DEVICE = 2
    FUNCTION_SET_ASSIGNMENT = 3
    RESPONSE = 4
    DEMAND_RESPONSE_LOAD_CONTROL = 5
    METERING = 6
    PRICING = 7
    MESSAGING = 8
    BILLING = 9
    PREPAYMENT = 10
    DISTRIBUTED_ENERGY_RESOURCES = 11
    TIME = 12
    SOFTWARE_DOWNLOAD = 13
    DEVICE_INFORMATION = 14
    POWER_STATUS = 15
    NETWORK_STATUS = 16
    LOG_EVENT_LIST = 17
    CONFIGURATION = 18
    SECURITY = 19


class ProfileIdentifier(IntEnum):
    """All other values are reserved"""

    NOT_PROFILE_SPECIFIC = 0
    VENDOR_DEFINED = 1
    IEEE_2030_5 = 2
    HOM_AUTOMATION = 3
    BUILDING_AUTOMATION = 4


class LogEvent(Resource):
    """A time stamped instance of a significant event detected by the device."""

    createdDateTime: types.TimeType = element()
    details: Optional[primitive_types.String32] = element(
        default=None
    )  # Human readable text that MAY be used to transmit additional details about the event. Host may remove this.

    extendedData: Optional[int] = element()  # May be used to transmit additional details about the event.
    functionSet: FunctionSetIdentifier = element()
    logEventCode: int = element()  # An 8 bit unsigned integer. logEventCodes are scoped to a profile and a function set
    logEventID: int = element()  # An 16 bit unsigned integer.
    logEventPEN: int = element()  # A 32 bit private enterprise number (PEN) of the entity creating this LogEvent
    profileID: ProfileIdentifier = element()  # An 8 bit unsigned integer.


class LogEventList(SubscribableList):
    pollRate: Optional[int] = attr(default=types.DEFAULT_POLLRATE_SECONDS)  # recommended client pollrate in seconds
    LogEvent_: Optional[list[LogEvent]] = element(default=None, tag="LogEvent")



from enum import IntEnum
from typing import Any, Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2.base import BaseXmlModelWithNS
from envoy_schema.server.schema.sep2.der import (
    AbnormalCategoryType,
    ConnectStatusTypeValue,
    DERControlBase,
    DERControlResponse,
    DERType,
    InverterStatusTypeValue,
    LocalControlModeStatusTypeValue,
    ManufacturerStatusValue,
    NormalCategoryType,
    OperationalModeStatusTypeValue,
    StateOfChargeStatusValue,
    StorageModeStatusTypeValue,
)
from envoy_schema.server.schema.sep2.der_control_types import (
    ActivePower,
    AmpereHour,
    ApparentPower,
    CurrentRMS,
    PowerFactor,
    ReactivePower,
    ReactiveSusceptance,
    VoltageRMS,
    WattHour,
)
from envoy_schema.server.schema.sep2.end_device import EndDeviceResponse
from envoy_schema.server.schema.sep2.identification import List as Sep2List
from envoy_schema.server.schema.sep2.identification import Resource
from envoy_schema.server.schema.sep2.metering import Reading
from envoy_schema.server.schema.sep2.pricing import TimeTariffIntervalResponse
from envoy_schema.server.schema.sep2.primitive_types import (
    HexBinary8,
    HexBinary32,
    HttpUri,
    LocalAbsoluteUri,
    HexBinary128,
)
from envoy_schema.server.schema.sep2.types import PerCent, SubscribableType, TimeType, VersionType

XSI_TYPE_TIME_TARIFF_INTERVAL_LIST = "TimeTariffIntervalList"
XSI_TYPE_DER_CONTROL_LIST = "DERControlList"
XSI_TYPE_DER_AVAILABILITY = "DERAvailability"
XSI_TYPE_DER_CAPABILITY = "DERCapability"
XSI_TYPE_DER_SETTINGS = "DERSettings"
XSI_TYPE_DER_STATUS = "DERStatus"
XSI_TYPE_DEFAULT_DER_CONTROL = "DefaultDERControl"
XSI_TYPE_END_DEVICE_LIST = "EndDeviceList"
XSI_TYPE_READING_LIST = "ReadingList"
XSI_TYPE_RESOURCE = "Resource"
XSI_TYPE_DEFAULT = XSI_TYPE_RESOURCE


class NotificationStatus(IntEnum):
    """Status values pertaining to Notification.status as described by Notification schema"""

    DEFAULT = 0
    SUBSCRIPTION_CANCELLED_NO_INFO = 1
    SUBSCRIPTION_CANCELLED_RESOURCE_MOVED = 2
    SUBSCRIPTION_CANCELLED_RESOURCE_DEFINITION_CHANGED = 3  # eg - new version of IEEE 2030.5
    SUBSCRIPTION_CANCELLED_RESOURCE_DELETED = 4


class SubscriptionEncoding(IntEnum):
    """Status values pertaining to Subscription.encoding as described by Subscription schema"""

    XML = 0  # application/sep+xml
    EXI = 1  # application/sep-exi


class ConditionAttributeIdentifier(IntEnum):
    """Status values pertaining to Condition.attributeIdentifier as described by Condition schema"""

    READING_VALUE = 0


class SubscriptionBase(Resource):
    """Holds the information related to a client subscription to receive updates to a resource automatically.
    The actual resources may be passed in the Notification by specifying a specific xsi:type for the Resource and
    passing the full representation."""

    subscribedResource: LocalAbsoluteUri = element()  # The resource for which the subscription applies.


def get_notification_resource_discriminator(v: Any) -> str:
    if not v:
        return XSI_TYPE_DEFAULT

    if isinstance(v, dict):
        return v.get("type", XSI_TYPE_DEFAULT)
    return getattr(v, "type", XSI_TYPE_DEFAULT)


class NotificationResourceCombined(Resource):
    """This class only exists because pydantic-xml has limited support for pydantic discriminated unions

    Ultimately we have a single element in notification called <Resource> that can be filled with any number
    of types - the pydantic xml typing struggles to represent this via discriminators.

    One major limitation: https://github.com/dapper91/pydantic-xml/issues/157
    Pydantic XML also only support discriminating between sub models with the exact same attributes

    This class essentially combines (manually) the following classes:
        TimeTariffIntervalListResponse, DERControlListResponse, DefaultDERControl, EndDeviceListResponse, Reading

    The plan is for the server to only fill out the fields relevant for the notification being served (based on
    the xsi:type attribute). Clients using this to parse Notifications will have to manually map the fields to
    the appropriate types.

    HERE BE DRAGONS FOR XSD VALIDITY:
        In order to keep XSD element ordering, we've had to do a few "creative" element orderings to ensure that
        generated notifications have elements in the XSD valid order. Because a couple of element names are shared
        between the combined resource types (eg setESDelay on DefaultDERControl / DERSettings) - We've had to define
        things in a confusing order - There isn't another way around this. See highlighted items marked with SORRY.
    """

    subscribable: Optional[SubscribableType] = attr(default=None)

    # List
    all_: Optional[int] = attr(name="all", default=None)
    results: Optional[int] = attr(default=None)

    # TimeTariffIntervalListResponse
    TimeTariffInterval: Optional[list[TimeTariffIntervalResponse]] = element(default=None)

    # DERControlListResponse
    DERControl: Optional[list[DERControlResponse]] = element(default=None)

    # EndDeviceListResponse
    EndDevice: Optional[list[EndDeviceResponse]] = element(default=None)

    # ReadingListResponse
    Readings: Optional[list[Reading]] = element(default=None, tag="Reading")

    # SubscribableIdentifiedObject
    mRID: Optional[HexBinary128] = element(default=None)
    description: Optional[str] = element(default=None)
    version: Optional[VersionType] = element(default=None)

    # SORRY (see docstring): DERSettings:  but because of the shared elements with DefaultDERControl, this must
    # appear above DefaultDERControl
    modesEnabled: Optional[HexBinary32] = element(default=None)  # SORRY

    # DefaultDERControl
    DERControlBase_: Optional[DERControlBase] = element(tag="DERControlBase", default=None)
    setESDelay: Optional[int] = element(default=None)
    setESHighFreq: Optional[int] = element(default=None)
    setESHighVolt: Optional[int] = element(default=None)
    setESLowFreq: Optional[int] = element(default=None)
    setESLowVolt: Optional[int] = element(default=None)
    setESRampTms: Optional[int] = element(default=None)
    setESRandomDelay: Optional[int] = element(default=None)
    setGradW: Optional[int] = element(default=None)
    # setSoftGradW: Optional[int] = element(default=None) # Duplicated from DERSettings

    # SORRY (see docstring): DERAvailability but unfortunately DERAvailability/DERStatus: both share readingTime, these
    # need to be brought up there to ensure they work if either resource type is populated
    availabilityDuration: Optional[int] = element(default=None)  # SORRY
    maxChargeDuration: Optional[int] = element(default=None)  # SORRY

    # DERStatus
    alarmStatus: Optional[HexBinary32] = element(default=None)
    genConnectStatus: Optional[ConnectStatusTypeValue] = element(default=None, tag="genConnectStatus")
    inverterStatus: Optional[InverterStatusTypeValue] = element(default=None, tag="inverterStatus")
    localControlModeStatus: Optional[LocalControlModeStatusTypeValue] = element(
        default=None, tag="localControlModeStatus"
    )
    manufacturerStatus: Optional[ManufacturerStatusValue] = element(default=None, tag="manufacturerStatus")
    operationalModeStatus: Optional[OperationalModeStatusTypeValue] = element(default=None, tag="operationalModeStatus")
    readingTime: Optional[TimeType] = element(default=None)
    stateOfChargeStatus: Optional[StateOfChargeStatusValue] = element(default=None, tag="stateOfChargeStatus")
    storageModeStatus: Optional[StorageModeStatusTypeValue] = element(default=None, tag="storageModeStatus")
    storConnectStatus: Optional[ConnectStatusTypeValue] = element(default=None, tag="storConnectStatus")

    # DERAvailability
    # readingTime: TimeType = element()  # Duplicated from DERStatus
    reserveChargePercent: Optional[PerCent] = element(default=None)
    reservePercent: Optional[PerCent] = element(default=None)
    statVarAvail: Optional[ReactivePower] = element(default=None)
    statWAvail: Optional[ActivePower] = element(default=None)

    # DERCapability
    modesSupported: Optional[HexBinary32] = element(default=None)
    rtgAbnormalCategory: Optional[AbnormalCategoryType] = element(default=None)
    rtgMaxA: Optional[CurrentRMS] = element(default=None)
    rtgMaxAh: Optional[AmpereHour] = element(default=None)
    rtgMaxChargeRateVA: Optional[ApparentPower] = element(default=None)
    rtgMaxChargeRateW: Optional[ActivePower] = element(default=None)
    rtgMaxDischargeRateVA: Optional[ApparentPower] = element(default=None)
    rtgMaxDischargeRateW: Optional[ActivePower] = element(default=None)
    rtgMaxV: Optional[VoltageRMS] = element(default=None)
    rtgMaxVA: Optional[ApparentPower] = element(default=None)
    rtgMaxVar: Optional[ReactivePower] = element(default=None)
    rtgMaxVarNeg: Optional[ReactivePower] = element(default=None)
    rtgMaxW: Optional[ActivePower] = element(default=None)
    rtgMaxWh: Optional[WattHour] = element(default=None)
    rtgMinPFOverExcited: Optional[PowerFactor] = element(default=None)
    rtgMinPFUnderExcited: Optional[PowerFactor] = element(default=None)
    rtgMinV: Optional[VoltageRMS] = element(default=None)
    rtgNormalCategory: Optional[NormalCategoryType] = element(default=None)
    rtgOverExcitedPF: Optional[PowerFactor] = element(default=None)
    rtgOverExcitedW: Optional[ActivePower] = element(default=None)
    rtgReactiveSusceptance: Optional[ReactiveSusceptance] = element(default=None)
    rtgUnderExcitedPF: Optional[PowerFactor] = element(default=None)
    rtgUnderExcitedW: Optional[ActivePower] = element(default=None)
    rtgVNom: Optional[VoltageRMS] = element(default=None)
    type_: Optional[DERType] = element(tag="type", default=None)
    doeModesSupported: Optional[HexBinary8] = element(ns="csipaus", default=None)

    # DERSettings
    # setESDelay: Optional[int] = element(default=None)  # Duplicated from DERControl
    # setESHighFreq: Optional[int] = element(default=None) # Duplicated from DERControl
    # setESHighVolt: Optional[int] = element(default=None) # Duplicated from DERControl
    # setESLowFreq: Optional[int] = element(default=None) # Duplicated from DERControl
    # setESLowVolt: Optional[int] = element(default=None) # Duplicated from DERControl
    # setESRampTms: Optional[int] = element(default=None) # Duplicated from DERControl
    # setESRandomDelay: Optional[int] = element(default=None) # Duplicated from DERControl
    # setGradW: int = element() # Duplicated from DERControl
    setMaxA: Optional[CurrentRMS] = element(default=None)
    setMaxAh: Optional[AmpereHour] = element(default=None)
    setMaxChargeRateVA: Optional[ApparentPower] = element(default=None)
    setMaxChargeRateW: Optional[ActivePower] = element(default=None)
    setMaxDischargeRateVA: Optional[ApparentPower] = element(default=None)
    setMaxDischargeRateW: Optional[ActivePower] = element(default=None)
    setMaxV: Optional[VoltageRMS] = element(default=None)
    setMaxVar: Optional[ReactivePower] = element(default=None)
    setMaxVarNeg: Optional[ReactivePower] = element(default=None)
    setMaxW: Optional[ActivePower] = element(default=None)
    setMaxWh: Optional[WattHour] = element(default=None)
    setMinPFOverExcited: Optional[PowerFactor] = element(default=None)
    setMinPFUnderExcited: Optional[PowerFactor] = element(default=None)
    setMinV: Optional[VoltageRMS] = element(default=None)
    setSoftGradW: Optional[int] = element(default=None)
    setVNom: Optional[VoltageRMS] = element(default=None)
    setVRef: Optional[VoltageRMS] = element(default=None)
    setVRefOfs: Optional[VoltageRMS] = element(default=None)
    updatedTime: Optional[TimeType] = element(default=None)
    doeModesEnabled: Optional[HexBinary8] = element(ns="csipaus", default=None)


class Notification(SubscriptionBase):
    """Holds the information related to a client subscription to receive updates to a resource automatically.
    The actual resources may be passed in the Notification by specifying a specific xsi:type for the Resource and
    passing the full representation."""

    newResourceURI: Optional[LocalAbsoluteUri] = element(default=None)  # The new location of the resource if moved.

    # A resource is an addressable unit of information, either a collection (List) or instance of an object
    # (identifiedObject, or simply, Resource)
    #
    # The xsi:type attribute will define how the entity is parsed
    #
    # NOTE - Resource must define an xsi:type attribute otherwise it will parse to Resource - logic is handled
    #      - in the pydantic Discriminator function get_notification_resource_discriminator
    #
    # NOTE - For more info - see pydantic docs on Unions / Discriminated Unions - Feature introduced in 2.5
    resource: Optional[
        # This callable discriminator union isn't supported by pydantic XML
        # see: https://github.com/dapper91/pydantic-xml/issues/157 - we might be able to swap to this in the future
        # Annotated[
        #     Union[
        #         Annotated[TimeTariffIntervalListResponse, Tag(XSI_TYPE_TIME_TARIFF_INTERVAL_LIST)],
        #         Annotated[DERControlListResponse, Tag(XSI_TYPE_DER_CONTROL_LIST)],
        #         Annotated[DefaultDERControl, Tag(XSI_TYPE_DEFAULT_DER_CONTROL)],
        #         Annotated[EndDeviceListResponse, Tag(XSI_TYPE_END_DEVICE_LIST)],
        #         Annotated[ReadingListResponse, Tag(XSI_TYPE_READING)],
        #         Annotated[Resource, Tag(XSI_TYPE_RESOURCE)],
        #     ],
        #     Discriminator(get_notification_resource_discriminator),
        # ]
        NotificationResourceCombined  # Instead we use this as our workaround for now
    ] = element(tag="Resource", default=None)
    status: NotificationStatus = element()
    subscriptionURI: LocalAbsoluteUri = element()  # Subscription from which this notification was triggered.


class Condition(BaseXmlModelWithNS):
    """Indicates a condition that must be satisfied for the Notification to be triggered."""

    attributeIdentifier: ConditionAttributeIdentifier = element()
    lowerThreshold: int = element(default=None)  # The value of the lower threshold
    upperThreshold: int = element(default=None)  # The value of the upper threshold


class Subscription(SubscriptionBase):
    """Holds the information related to a client subscription to receive updates to a resource automatically."""

    condition: Optional[Condition] = element(tag="Condition", default=None)
    encoding: SubscriptionEncoding = element()  # The resource for which the subscription applies.
    level: str = element()  # Contains the preferred schema and extensibility level indication such as "+S1"
    limit: int = element()  # This element is used to indicate the maximum number of list items that should be included
    # in a notification when the subscribed resource changes
    notificationURI: HttpUri = element()  # The resource to which to post the notifications


class SubscriptionListResponse(Sep2List, tag="SubscriptionList"):
    subscriptions: Optional[list[Subscription]] = element(tag="Subscription", default=None)
    pollRate: Optional[int] = attr(default=None)  # The default polling rate for this function set in seconds


class NotificationListResponse(Sep2List, tag="NotificationList"):
    notifications: Optional[list[Notification]] = element(tag="Notification", default=None)



from datetime import datetime
from decimal import Decimal
from typing import Optional

from pydantic import BaseModel


class DynamicOperatingEnvelopeRequest(BaseModel):
    """Dynamic Operating Envelope basic model"""

    site_id: int  # Corresponds to EndDevice id
    calculation_log_id: Optional[int]  # The ID of the CalculationLog that created this DOE (or NULL if no link)
    duration_seconds: int
    import_limit_active_watts: Decimal
    export_limit_watts: Decimal
    start_time: datetime


class DynamicOperatingEnvelopeResponse(DynamicOperatingEnvelopeRequest):
    """Dynamic Operating Envelope basic model when being queried externally"""

    dynamic_operating_envelope_id: int  # Internal identifier for this DOE
    created_time: datetime
    changed_time: datetime  # When this DOE was last changed/created


class DoePageResponse(BaseModel):
    """Represents a paginated response of DynamicOperatingEnvelopeRequest"""

    total_count: int  # The total number of does (independent of this page of results)
    limit: int  # The maximum number of does that could've been returned (the limit set by the query)
    start: int  # The number of does that have been skipped as part of this query (the start set by the query)
    after: Optional[datetime]  # The "after" filter set by the query
    does: list[DynamicOperatingEnvelopeResponse]  # The doe models in this paged response



from typing import Optional

from pydantic_xml import element

from envoy_schema.server.schema.sep2 import base, types


class ErrorResponse(base.BaseXmlModelWithNS, tag="Error"):
    """Represents a description of a request error and how the client should respond"""

    maxRetryDuration: Optional[int] = element(
        default=None
    )  # Contains the number of seconds the client SHOULD wait before retrying
    reasonCode: types.ReasonCodeType = element()  # Code indicating the reason for failure.

    # These properties sit outside the sep2 definition and are our own custom extensions to provide clients
    # with some QoL improvements
    message: Optional[str] = element(default=None)



from enum import IntEnum
from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2 import primitive_types
from envoy_schema.server.schema.sep2.base import BaseXmlModelWithNS
from envoy_schema.server.schema.sep2.identification import IdentifiedObject
from envoy_schema.server.schema.sep2.identification import List as Sep2List
from envoy_schema.server.schema.sep2.identification import ListLink, Resource
from envoy_schema.server.schema.sep2.time import TimeType
from envoy_schema.server.schema.sep2.types import DEFAULT_POLLRATE_SECONDS, PerCent


class ResponseType(IntEnum):
    """Derived from sep2 Table 27 - Response types by function set. All other values reserved"""

    EVENT_RECEIVED = 1
    EVENT_STARTED = 2
    EVENT_COMPLETED = 3
    USER_CHOSE_OPT_OUT = 4
    USER_CHOSE_OPT_IN = 5
    EVENT_CANCELLED = 6
    EVENT_SUPERSEDED = 7
    EVENT_PARTIAL_COMPLETE_USER_OPT_OUT = 8
    EVENT_PARTIAL_COMPLETE_USER_OPT_IN = 9
    EVENT_COMPLETE_NO_USER_PARTICIPATION = 10
    USER_ACKNOWLEDGED_EVENT = 11
    CANNOT_BE_DISPLAYED = 12
    EVENT_ABORTED_DUE_TO_ALTERNATE_SERVER_EVENT = 13
    EVENT_ABORTED_DUE_TO_ALTERNATE_PROGRAM_EVENT = 14
    REJECTED_CONTROL_PARAM_NOT_APPLICABLE = 252
    REJECTED_INVALID_EVENT = 253
    REJECTED_EVENT_RECEIVED_LATE = 254


class ApplianceLoadReductionType(IntEnum):
    """Full definition of how appliances react when receiving each parameter is document in the EPA document,
    ENERGY STAR® Program Requirements, Product Specification for Residential Refrigerators and
    Freezers, Eligibility Criteria 5, Draft 2 Version 5.0.
    All other values reserved"""

    DELAY_APPLIANCE_LOAD = 0
    TEMPORARY_APPLIANCE_LOAD_REDUCTION = 1


class ResponseSet(IdentifiedObject):
    """A container for a ResponseList"""

    ResponseListLink: Optional[ListLink] = element(default=None)


class Response(Resource):
    """The Response object is the generic response data repository which is extended for specific function sets."""

    createdDateTime: Optional[TimeType] = element(default=None)
    endDeviceLFDI: primitive_types.HexBinary160 = element()
    status: Optional[ResponseType] = element(default=None)
    subject: primitive_types.HexBinary128 = element()


class ResponseListResponse(Sep2List, tag="ResponseList"):
    """List element for holding Response objects"""

    Response_: Optional[list[Response]] = element(default=None, tag="Response")


class ResponseSetList(Sep2List):
    """A List element to hold ResponseSet objects."""

    pollRate: Optional[int] = attr(default=DEFAULT_POLLRATE_SECONDS)  # recommended client pollrate in seconds
    ResponseSet_: Optional[list[ResponseSet]] = element(default=None, tag="ResponseSet")


class DERControlResponse(Response):
    """A response to a DERControl"""

    pass


class FlowReservationResponseResponse(Response):
    """A response to a FlowReservationResponse"""

    pass


class PriceResponse(Response):
    """A response related to a price message."""

    pass


class TextResponse(Response):
    """A response related to a text message."""

    pass


class ApplianceLoadReduction(BaseXmlModelWithNS):
    """The ApplianceLoadReduction object is used by a Demand Response service provider to provide signals for ENERGY
    STAR compliant appliances. See the definition of ApplianceLoadReductionType for more information."""

    type: int = element()  # Values drawn from ApplianceLoadReductionType


class AppliedTargetReduction(BaseXmlModelWithNS):
    """Specifies the value of the TargetReduction applied by the device."""

    type: int = element()  # Values drawn from UnitValueType
    value: int = element()  # The requested amount of the relevant commodity to be reduced.


class SetPoint(BaseXmlModelWithNS):
    """The SetPoint object is used to apply specific temperature set points to a temperature control device. The values
    of the heatingSetpoint and coolingSetpoint attributes SHALL be calculated as follows:

    Cooling/Heating Temperature Set Point / 100 = temperature in degrees Celsius where
    -273.15°C <= temperature <= 327.67°C, corresponding to a Cooling and/or Heating Temperature Set Point.

    The maximum resolution this format allows is 0.01°C.

    The field not present in a Response indicates that this field has not been used by the end device.
    If a temperature is sent that exceeds the temperature limit boundaries that are programmed into the device,
    the device SHALL respond by setting the temperature at the limit."""

    coolingSetpoint: int = (
        element()
    )  # This attribute represents the cooling temp set point in degrees Celsius / 100. (Hundredths of a degree C)

    heatingSetpoint: int = (
        element()
    )  # This attribute represents the heating temp set point in degrees Celsius / 100. (Hundredths of a degree C)


class Offset(BaseXmlModelWithNS):
    """If a temperature offset is sent that causes the heating or cooling temperature set point to exceed the limit
    boundaries that are programmed into the device, the device SHALL respond by setting the temperature at the limit.

    If an EDC is being targeted at multiple devices or to a device that controls multiple devices (e.g., EMS), it can
    provide multiple Offset types within one EDC. For events with multiple Offset types, a client SHALL select the
    Offset that best fits their operating function.

    Alternatively, an event with a single Offset type can be targeted at an EMS in order to request a percentage load
    reduction on the average energy usage of the entire premise. An EMS SHOULD use the Metering function set to
    determine the initial load in the premise, reduce energy consumption by controlling devices at its disposal, and
    at the conclusion of the event, once again use the Metering function set to determine if the desired load reduction
    was achieved."""

    # The value change requested for the cooling offset, in degree C / 10. The value should be added to the normal set
    # point for cooling, or if loadShiftForward is true, then the value should be subtracted from the normal set point.
    coolingOffset: Optional[int] = element(default=None)

    # The value change requested for the heating offset, in degree C / 10. The value should be subtracted for heating,
    # or if loadShiftForward is true, then the value should be added to the normal set point.
    heatingOffset: Optional[int] = element(default=None)

    # The value change requested for the load adjustment percentage. The value should be subtracted from the normal
    # setting, or if loadShiftForward is true, then the value should be added to the normal setting.
    loadAdjustmentPercentageOffset: Optional[PerCent] = element(default=None)


class DutyCycle(BaseXmlModelWithNS):
    """Duty cycle control is a device specific issue and is managed by the device.  The duty cycle of the device under
    control should span the shortest practical time period in accordance with the nature of the device under control
    and the intent of the request for demand reduction.

    The default factory setting SHOULD be three minutes for each 10% of duty cycle.  This indicates that the default
    time period over which a duty cycle is applied is 30 minutes, meaning a 10% duty cycle would cause a device to be
    ON for 3 minutes.

    The “off state” SHALL precede the “on state”."""

    # Contains the maximum On state duty cycle applied by the end device, as a percentage of time.  The field not
    # present indicates that this field has not been used by the end device.
    normalValue: int = element()


class DrResponse(Response):
    """A response to a Demand Response Load Control (EndDeviceControl) message."""

    ApplianceLoadReduction_: Optional[ApplianceLoadReduction] = element(default=None, tag="ApplianceLoadReduction")

    AppliedTargetReduction_: Optional[AppliedTargetReduction] = element(default=None, tag="AppliedTargetReduction")
    DutyCycle_: Optional[DutyCycle] = element(default=None, tag="DutyCycle")
    Offset_: Optional[Offset] = element(default=None, tag="Offset")
    # Indicates the amount of time, in seconds, that the client partially opts-out during the demand response event.
    # When overriding within the allowed override duration, the client SHALL send a partial opt-out (Response status
    # code 8) for partial opt-out upon completion, with the total time the event was overridden (this attribute)
    # populated. The client SHALL send a no participation status response (status type 10) if the user partially
    # opts-out for longer than EndDeviceControl.overrideDuration.
    overrideDuration: Optional[int] = element(default=None)
    SetPoint_: Optional[SetPoint] = element(default=None, tag="SetPoint")



from typing import Optional

from pydantic_xml import element

from envoy_schema.server.schema.sep2 import base
from envoy_schema.server.schema.sep2.identification import Link


class ConnectionPointLink(Link, ns="csipaus"):
    pass


class ConnectionPointRequest(base.BaseXmlModelWithNS, tag="ConnectionPoint", ns="csipaus"):
    """Contains identification information related to the network location at which the EndDevice is installed."""

    id_v11: Optional[str] = element(default=None, tag="id")  # Typically used as the NMI (legacy version for csip1.1)
    id: Optional[str] = element(default=None, tag="connectionPointId")  # Typically used as the NMI (valid from 1.1a)


class ConnectionPointResponse(base.BaseXmlModelWithNS, tag="ConnectionPoint", ns="csipaus"):
    """Contains identification information related to the network location at which the EndDevice is installed."""

    id: str = element(default=None, tag="connectionPointId")  # Typically used as the NMI (valid from 1.1a)



"""Defines all the URIs"""

AggregatorListUri = "/aggregator"
AggregatorUri = "/aggregator/{aggregator_id}"
ArchiveForPeriodSites = "/archive/{period_start}/{period_end}/sites"
ArchiveForPeriodDoes = "/archive/{period_start}/{period_end}/does"
ArchiveForPeriodTariffGeneratedRate = "/archive/{period_start}/{period_end}/tariff_generated_rates"
TariffCreateUri = "/tariff"
TariffUpdateUri = "/tariff/{tariff_id}"
TariffGeneratedRateCreateUri = "/tariff_generated_rate"
DoeUri = "/doe"
SiteUri = "/site"
SiteGroupUri = "/site_group/{group_name}"
SiteGroupListUri = "/site_group"
AggregatorBillingUri = "/billing/aggregator/{aggregator_id}/tariff/{tariff_id}/period/{period_start}/{period_end}"
SitePeriodBillingUri = "/billing/site/period"
CalculationLogBillingUri = "/billing/calculation_log/{calculation_log_id}/tariff/{tariff_id}"
CalculationLogCreateUri = "/calculation_log"
CalculationLogUri = "/calculation_log/{calculation_log_id}"
CalculationLogsForPeriod = "/calculation_log/period/{period_start}/{period_end}"



from datetime import datetime
from decimal import Decimal
from typing import Optional

from pydantic import BaseModel

from envoy_schema.server.schema.sep2.types import CurrencyCode


class TariffRequest(BaseModel):
    """Basic attributes for the creation of a new tariff structure."""

    name: str
    dnsp_code: str
    currency_code: CurrencyCode


class TariffResponse(BaseModel):
    """Response model for Tariff including id and modification time."""

    tariff_id: int
    name: str
    dnsp_code: str
    currency_code: CurrencyCode
    created_time: datetime
    changed_time: datetime


class TariffGeneratedRateRequest(BaseModel):
    """Time of use tariff pricing - represents a price for a specific site for a specific range of time"""

    tariff_id: int
    site_id: int
    calculation_log_id: Optional[int]  # The ID of the CalculationLog that created this rate (or NULL if no link)
    start_time: datetime
    duration_seconds: int
    import_active_price: Decimal  # Price in dollars per kw/h (holds 4 decimal places of precision)
    export_active_price: Decimal  # Price in dollars per kw/h (holds 4 decimal places of precision)
    import_reactive_price: Decimal  # Price is dollars per kvar/h (holds 4 decimal places of precision)
    export_reactive_price: Decimal  # Price is dollars per kvar/h (holds 4 decimal places of precision)


class TariffGeneratedRateResponse(TariffGeneratedRateRequest):
    tariff_generated_rate_id: int
    created_time: datetime
    changed_time: datetime



from typing import Optional

from pydantic_xml import element

from envoy_schema.server.schema.sep2.event import RandomizableEvent
from envoy_schema.server.schema.sep2.identification import IdentifiedObject, Link
from envoy_schema.server.schema.sep2.identification import List as SepList
from envoy_schema.server.schema.sep2.identification import ListLink, Resource, SubscribableList
from envoy_schema.server.schema.sep2.types import (
    ConsumptionBlockType,
    CurrencyCode,
    PrimacyType,
    ServiceKind,
    TOUType,
    UnitValueType,
)
from envoy_schema.server.schema.sep2.primitive_types import HexBinary16


class TariffProfileResponse(IdentifiedObject, tag="TariffProfile"):
    """A schedule of charges; structure that allows the definition of tariff structures such as step (block) and
    time of use (tier) when used in conjunction with TimeTariffInterval and ConsumptionTariffInterval."""

    currency: Optional[CurrencyCode] = element(default=None)
    pricePowerOfTenMultiplier: Optional[int] = element(default=None)
    primacyType: PrimacyType = element(default=None, tag="primacy")
    rateCode: Optional[str] = element(default=None)
    RateComponentListLink: Optional[ListLink] = element(default=None)
    serviceCategoryKind: ServiceKind = element()


class RateComponentResponse(IdentifiedObject, tag="RateComponent"):
    """Specifies the applicable charges for a single component of the rate, which could be generation price or
    consumption price, for example."""

    ActiveTimeTariffIntervalListLink: Optional[ListLink] = element(default=None)
    flowRateEndLimit: Optional[UnitValueType] = element(default=None)
    flowRateStartLimit: Optional[UnitValueType] = element(default=None)
    ReadingTypeLink: Link = element()
    roleFlags: HexBinary16 = element()  # See RoleFlagsType

    TimeTariffIntervalListLink: ListLink = element()


class TimeTariffIntervalResponse(RandomizableEvent, tag="TimeTariffInterval"):
    """Describes the time-differentiated portion of the RateComponent, if applicable, and provides the ability to
    specify multiple time intervals, each with its own consumption-based components and other attributes."""

    ConsumptionTariffIntervalListLink: ListLink = element()
    touTier: TOUType = element()


class ConsumptionTariffIntervalResponse(Resource, tag="ConsumptionTariffInterval"):
    """One of a sequence of thresholds defined in terms of consumption quantity of a service such as electricity,
    water, gas, etc. It defines the steps or blocks in a step tariff structure, where startValue simultaneously
    defines the entry value of this step and the closing value of the previous step. Where consumption is greater
    than startValue, it falls within this block and where consumption is less than or equal to startValue, it falls
    within one of the previous blocks."""

    consumptionBlock: ConsumptionBlockType = element()
    price: Optional[int] = element(
        default=None
    )  # The charge for this rate component, per unit of measure defined by the
    # associated ReadingType, in currency specified in TariffProfile.  # noqa e114
    startValue: int = element()  # The lowest level of consumption that defines the starting point of this consumption
    # step or block. Thresholds start at zero for each billing period. # noqa e114


class TariffProfileListResponse(SepList, tag="TariffProfileList"):
    TariffProfile: Optional[list[TariffProfileResponse]] = element(default=None)


class RateComponentListResponse(SubscribableList, tag="RateComponentList"):
    """Worth noting that the standard describes RateComponentList as a standard list but it's an envoy
    specific extension to support subscriptions via SubscribableList"""

    RateComponent: Optional[list[RateComponentResponse]] = element(default=None)


class TimeTariffIntervalListResponse(SepList, tag="TimeTariffIntervalList"):
    TimeTariffInterval: Optional[list[TimeTariffIntervalResponse]] = element(default=None)


class ConsumptionTariffIntervalListResponse(SepList, tag="ConsumptionTariffIntervalList"):
    ConsumptionTariffInterval: Optional[list[ConsumptionTariffIntervalResponse]] = element(default=None)



from pydantic_xml import BaseXmlModel
from pydantic_xml.element import SearchMode

nsmap = {
    "": "urn:ieee:std:2030.5:ns",
    "csipaus": "https://csipaus.org/ns",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
}


class BaseXmlModelWithNS(BaseXmlModel):
    model_config = {"arbitrary_types_allowed": True}

    def __init_subclass__(
        cls,
        *args,
        **kwargs,
    ):
        super().__init_subclass__(*args, **kwargs)
        cls.__xml_nsmap__ = nsmap
        cls.__xml_search_mode__ = SearchMode.UNORDERED



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class CalculationLogVariableMetadata(BaseModel):
    """This is purely descriptive metadata to highlight what an opaque CalculationLogVariable represents"""

    variable_id: int  # The variable that this metadata applies to
    name: str  # Short, human readable name for the referenced variable
    description: str  # Longer, human readable description of the variable


class CalculationLogLabelMetadata(BaseModel):
    """This is purely descriptive metadata to highlight what an opaque CalculationLogLabel represents"""

    label_id: int  # The variable that this metadata applies to
    name: str  # Short, human readable name for the referenced variable
    description: str  # Longer, human readable description of the variable


class CalculationLogVariableValues(BaseModel):
    """This is a compact representation of MANY variable instances. It's expected that EVERY property list has the
    same number of elements (i.e. there is a 1-1 correspondence between lists).

    eg: {
      variable_ids: [1,2]
      site_ids: [3, None]
      interval_periods: [5, 6]
      values: [7.7, 8.8]
    }

    Logically represents:

    [
        {variable_id: 1, site_id: 3, interval_period: 5, value: 7.7},
        {variable_id: 2, site_id: None, interval_period: 6, value: 8.8},
    ]

    A single variable value represents a single time series observation for a specific variable. The definition
    of what a variable represents is opaque to the utility server.

    Within a CalculationLog, the combination of variable_id, site_id and interval_period are a unique index


    The values (when returned from the server) will have the following sort order applied:
        variable_id ASCENDING
        site_id_snapshot ASCENDING
        interval_period ASCENDING
    """

    # ID defined by the client that disambiguate one set of time-series from another data from another. eg: a value of 1
    # might represent weather forecast temperature, a value of 2 might represent forecast load etc. The actual
    # definitions are completely opaque to utility server.
    variable_ids: list[int]  # Must correspond 1-1 with each other list in this type

    # Foreign key reference to a specific site ID that the variable value applies to or NONE if this time series
    # observation is NOT tied to a specific site.
    site_ids: list[Optional[int]]  # Must correspond 1-1 with each other list in this type

    # When does this time series observation occur? Defines the numbered "interval" relative to the parent
    # CalculationLog.calculation_range_start. A value of N uses the following formula for calculating datetime:
    # CalculationLog.calculation_range_start + N * CalculationLog.interval_width_seconds
    interval_periods: list[int]  # Must correspond 1-1 with each other list in this type

    # The actual time series value associated with the linked variable_id, site_id and interval_period
    values: list[float]  # Must correspond 1-1 with each other list in this type


class CalculationLogLabelValues(BaseModel):
    """This is a compact representation of MANY label instances. It's expected that EVERY property list has the
    same number of elements (i.e. there is a 1-1 correspondence between lists).

    eg: {
      label_ids: [1,2]
      site_ids: [3, None]
      values: ["Label 1", "Label 2"]
    }

    Logically represents:

    [
        {label_id: 1, site_id: 3, value: "Label 1"},
        {label_id: 2, site_id: None, value: "Label 2"},
    ]

    A single label value represents a free text descriptor for tagging a site/log with some extra information. The
    definition of what the label represents is opaque to the utility server.

    Within a CalculationLog, the combination of label_id and site_id are a unique index

    The values (when returned from the server) will have the following sort order applied:
        variable_id ASCENDING
        site_id_snapshot ASCENDING
    """

    # ID defined by the client that disambiguate one set of labels from another. eg: a value of 1
    # might represent an upstream transformer ID and a value of 2 might represent the name of a cohort that
    # dictates how the site was treated during this calculation run.
    # The actual definitions are completely opaque to utility server.
    label_ids: list[int]  # Must correspond 1-1 with each other list in this type

    # Foreign key reference to a specific site ID that the label applies to or NONE if this label is NOT tied to a
    # specific site.
    site_ids: list[Optional[int]]  # Must correspond 1-1 with each other list in this type

    # The actual label values associated with the linked label_id and site_id
    values: list[str]  # Must correspond 1-1 with each other list in this type


class CalculationLogRequest(BaseModel):
    """Represents the top level entity describing a single audit log of a historical calculation run.

    Calculation runs typically represent running powerflow / other model for some network. A calculation log represents
    a (mostly) opaque log of values defined by an external calculation engine. Any given calculation log has the
    following assumptions:
        * A calculation log represents a defined "range" of time for which the output calculations apply for
           eg: A single log might represent a 24 hour period of time - typically this range is in advance of when the
               calculations are being made.
        * A calculation log is divided into fixed width intervals of a known size, eg 5 minutes. All input data/outputs
          are aligned with these intervals. Eg - A 24 hour period is broken down into intervals of length 1 hour.
        * A calculation log has logged "variable" data representing input/intermediate/output data. This data is opaque
          to the utility server but it WILL align with intervals.
        * A calculation log has free-text "label" data associated with sites. This data is opaque to the utility server
          and is NOT time varying.
    """

    calculation_range_start: datetime  # The start time of the first interval within this calculation log.
    calculation_range_duration_seconds: int  # Number of seconds that define the width of this entire calculation log
    interval_width_seconds: int  # Number of seconds for the fixed width intervals that comprise this calculation log

    topology_id: Optional[str] = None  # The ID of the network topology being forecast (eg feeder ID)
    external_id: Optional[str] = None  # An ID for the external submitting client to identify this calculation log
    description: Optional[str] = None  # A human readable description of this calculation log

    power_forecast_creation_time: Optional[datetime] = None  # Datetime for when any power forecast was created

    # When was the last (most recent) historical lag. The time between this and the calculation_range_start represents
    # how stale the lag data was.
    power_forecast_basis_time: Optional[datetime] = None

    weather_forecast_creation_time: Optional[datetime] = None  # Datetime for when any weather forecast was created
    weather_forecast_location_id: Optional[str] = None  # ID associated the weather location that the forecast is for

    variable_metadata: list[
        CalculationLogVariableMetadata
    ]  # Metadata associated with the variables defined in variable_values

    # The actual time series observations in this calculation log
    # The values will have a defined sort order (see docs on CalculationLogVariableValues)
    variable_values: Optional[CalculationLogVariableValues]

    label_metadata: list[CalculationLogLabelMetadata]  # Metadata associated with the labels defined in label_values

    # The actual labels in this calculation log
    # The labels will have a defined sort order (see docs on CalculationLogLabelValues)
    label_values: Optional[CalculationLogLabelValues]


class CalculationLogResponse(CalculationLogRequest):
    """Represents the top level entity describing a single audit log of a historical calculation run.

    Calculation runs typically represent running powerflow / other model for some network based on forecast
    power/weather data (usually over multiple time steps) that may propose certain changes in DER behavior
    in order to satisfy certain network constraints"""

    calculation_log_id: int
    created_time: datetime


class CalculationLogListResponse(BaseModel):
    """A list response of CalculationLogResponse entities. Don't expect CalculationLogResponse to include
    any child logs - those can be fetched using calculation_log_id in a subsequent request.

    Represents a page of items and a total of all available items"""

    start: int
    limit: int
    total_calculation_logs: int
    calculation_logs: list[CalculationLogResponse]



from typing import Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.csip_aus.connection_point import ConnectionPointLink as ConnectionPointLinkType
from envoy_schema.server.schema.sep2 import primitive_types
from envoy_schema.server.schema.sep2.identification import (
    Link,
    ListLink,
    Resource,
    SubscribableList,
    SubscribableResource,
)
from envoy_schema.server.schema.sep2.time import TimeType
from envoy_schema.server.schema.sep2.types import DEFAULT_POLLRATE_SECONDS, PINType


class AbstractDevice(SubscribableResource):

    ConfigurationLink: Optional[Link] = element(default=None)
    DERListLink: Optional[ListLink] = element(default=None)
    deviceCategory: Optional[primitive_types.HexBinary32] = element(default=None)
    DeviceInformationLink: Optional[Link] = element(default=None)
    DeviceStatusLink: Optional[Link] = element(default=None)
    FileStatusLink: Optional[Link] = element(default=None)
    IPInterfaceListLink: Optional[Link] = element(default=None)
    lFDI: Optional[str] = element(default=None)
    LoadShedAvailabilityListLink: Optional[ListLink] = element(default=None)
    LogEventListLink: Optional[ListLink] = element(default=None)
    PowerStatusLink: Optional[Link] = element(default=None)
    sFDI: int = element()


class EndDeviceRequest(AbstractDevice, tag="EndDevice"):
    postRate: Optional[int] = element(default=None)


class EndDeviceResponse(AbstractDevice, tag="EndDevice"):
    changedTime: TimeType = element()
    enabled: Optional[bool] = element(default=True)

    FlowReservationRequestListLink: Optional[Link] = element(default=None)
    FlowReservationResponseListLink: Optional[Link] = element(default=None)
    FunctionSetAssignmentsListLink: Optional[ListLink] = element(default=None)
    postRate: Optional[int] = element(default=None)
    RegistrationLink: Optional[Link] = element(default=None)
    SubscriptionListLink: Optional[ListLink] = element(default=None)

    # csip extension
    ConnectionPointLink: Optional[ConnectionPointLinkType] = element(ns="csipaus", default=None)


class EndDeviceListResponse(SubscribableList, tag="EndDeviceList"):
    EndDevice: Optional[list[EndDeviceResponse]] = element(default=None)


class RegistrationResponse(Resource, tag="Registration"):
    pollRate: Optional[int] = attr(default=DEFAULT_POLLRATE_SECONDS)  # recommended client pollrate in seconds
    dateTimeRegistered: TimeType = element()  # Contains the time at which this registration was created
    pIN: PINType = (
        element()
    )  # Contains the registration PIN number associated with the device, includes checksum digit.



from typing import List, Optional

from pydantic_xml import attr, element

from envoy_schema.server.schema.sep2 import primitive_types, types
from envoy_schema.server.schema.sep2.identification import IdentifiedObject
from envoy_schema.server.schema.sep2.identification import List as Sep2List
from envoy_schema.server.schema.sep2.identification import Resource
from envoy_schema.server.schema.sep2.metering import Reading, ReadingSetBase, ReadingType, UsagePointBase


class MirrorReadingSet(ReadingSetBase):
    readings: Optional[List[Reading]] = element(tag="Reading", default=None)


class MeterReadingBase(IdentifiedObject):
    pass


class MirrorMeterReading(MeterReadingBase):
    lastUpdateTime: Optional[types.TimeType] = element(default=None)
    mirrorReadingSets: Optional[List[MirrorReadingSet]] = element(tag="MirrorReadingSet", default=None)
    nextUpdateTime: Optional[types.TimeType] = element(default=None)
    reading: Optional[Reading] = element(tag="Reading", default=None)
    readingType: Optional[ReadingType] = element(tag="ReadingType", default=None)


class MirrorUsagePoint(UsagePointBase):
    deviceLFDI: primitive_types.HexBinary160 = element()
    mirrorMeterReadings: Optional[List[MirrorMeterReading]] = element(tag="MirrorMeterReading", default=None)
    postRate: Optional[int] = element(default=None)


class MirrorUsagePointList(Sep2List):
    pollRate: Optional[int] = attr(default=types.DEFAULT_POLLRATE_SECONDS)  # recommended client pollrate in seconds
    mirrorUsagePoints: Optional[List[MirrorUsagePoint]] = element(tag="MirrorUsagePoint", default=None)


class MirrorMeterReadingList(Sep2List):
    mirrorMeterReadings: Optional[List[MirrorMeterReading]] = element(tag="MirrorMeterReading", default=None)


class MirrorMeterReadingRequest(MirrorMeterReading, tag="MirrorMeterReading"):
    pass


# Unlike MirrorMeterReadingList this is a list resource is doesn't subclass Sep2List. The reasons is that clients
# don't need to specify the attributes 'all' or 'result' for the list resources they are posting
class MirrorMeterReadingListRequest(Resource, tag="MirrorMeterReadingList"):
    mirrorMeterReadings: Optional[List[MirrorMeterReading]] = element(tag="MirrorMeterReading", default=None)


class MirrorUsagePointListResponse(Sep2List, tag="MirrorUsagePointList"):
    mirrorUsagePoints: list[MirrorUsagePoint] = element(tag="MirrorUsagePoint", default_factory=list)


class MirrorUsagePointRequest(MirrorUsagePoint, tag="MirrorUsagePoint"):
    pass
